/*********************************************************************************************
//
// OpenCL Histogram kernels for GPU
// 
// Based on https://developer.apple.com/library/mac/samplecode/gpu_histogram/Introduction/Intro.html
// Modifications:
// - Add additional bins for out-of-range values
// - Add luma component (always scaled to [0-255] + out of range values)
// - Support additional pixelformats
//
*********************************************************************************************/
#pragma OPENCL EXTENSION cl_khr_local_int32_base_atomics : enable
#define NUM_BINS 512
#define NUM_CHANNELS 4

//
// sum partial histogram results into final histogram bins
//
// num_groups is the number of work-groups used to compute partial histograms.
// partial_histogram is an array of num_groups * (256 * 3 * 32-bits/entry) entries
// we store 256 Red bins, followed by 256 Green bins and then the 256 Blue bins.
//
// final summed results are returned in histogram.
//
kernel
void histogram_sum_partial_results(global uint *partial_histogram, int num_groups, global uint *histogram)
{
    int     tid = (int)get_global_id(0);
    int     group_indx;
    int     n = num_groups;
    local uint  tmp_histogram[NUM_BINS * NUM_CHANNELS];

    tmp_histogram[tid] = partial_histogram[tid];
    
    group_indx = NUM_BINS*NUM_CHANNELS;
    while (--n > 0)
    {
        tmp_histogram[tid] += partial_histogram[group_indx + tid];
        group_indx += NUM_BINS*NUM_CHANNELS;
    }
    
    histogram[tid] = tmp_histogram[tid];
}

//
// this kernel takes an 8-bit / channel input image and produces a partial histogram.
// the kernel is executed over multiple work-groups.  for each work-group a partial histogram is generated
// partial_histogram is an array of num_groups * (NUM_BINS * 3 * 32-bits/entry) entries
// we store NUM_BINS Red bins, followed by NUM_BINS Green bins and then the NUM_BINS Blue bins.
//
kernel
#ifdef nv12_nv21
kernel void histogram(constant  float*    coef,
                                float     lumaZero,
                                float     chromaZero,
                                float     blackLevel,
                                float     depthScale,
                      global    uint*     histogram_buffer,
                      read_only image2d_t inPlane0,
                      read_only image2d_t inPlane1) {
#endif
#ifdef yuv_planar
kernel void histogram(constant  float*    coef,
                                float     lumaZero,
                                float     chromaZero,
                                float     blackLevel,
                                float     depthScale,
                      global    uint*     histogram_buffer,
                      read_only image2d_t inPlane0,
                      read_only image2d_t inPlane1,
                      read_only image2d_t inPlane2) {
#endif
#if defined(uyvy) || defined(_rgb)
kernel void histogram(constant  float*    coef,
                                float     lumaZero,
                                float     chromaZero,
                                float     blackLevel,
                                float     depthScale,
                      global    uint*     histogram_buffer,
                      read_only image2d_t inImage) {
#endif
    int     local_size = (int)get_local_size(0) * (int)get_local_size(1);
    int     group_indx = mad24(convert_uint(get_group_id(1)), convert_uint(get_num_groups(0)), convert_uint(get_group_id(0))) * NUM_BINS * NUM_CHANNELS;
    int     x = get_global_id(0);
    int     y = get_global_id(1);
    
    local uint  tmp_histogram[NUM_BINS * NUM_CHANNELS];
        
    int     tid = mad24(convert_uint(get_local_id(1)), convert_uint(get_local_size(0)), convert_uint(get_local_id(0)));
    int     j = NUM_BINS * NUM_CHANNELS;
    int     indx = 0;
    
    // clear the local buffer that will generate the partial histogram
    do
    {
        if (tid < j)
            tmp_histogram[indx+tid] = 0;

        j -= local_size;
        indx += local_size;
    } while (j > 0);
    
    barrier(CLK_LOCAL_MEM_FENCE);
    
    int2   coord  = (int2)(x, y);

    #ifdef nv12_nv21
    float2 coordIn = norm(coord, inPlane0);
    float3 rgb     = nv12_nv21_to_rgb(coef, lumaZero, chromaZero, blackLevel, depthScale, coordIn, inPlane0, inPlane1);
    float  luma    = read_imagef(inPlane0, sampler, coordIn).r * depthScale * 255.0f/219.0f - 16.0f/219.0f;
    #endif
    #ifdef uyvy
    float2 coordIn = norm(coord, inImage);
    float3 rgb     = uyvy_to_rgb(coef, lumaZero, chromaZero, blackLevel, depthScale, coordIn, inImage);
    float  luma    = dot((float3)(0.2126f, 0.7152f, 0.0722f), rgb);
    #endif
    #ifdef yuv_planar
    float2 coordIn = norm(coord, inPlane0);
    float3 rgb     = planar_yuv_to_rgb(coef, lumaZero, chromaZero, blackLevel, depthScale, coordIn, inPlane0, inPlane1, inPlane2);
    float  luma    = read_imagef(inPlane0, sampler, coordIn).r * depthScale * 255.0f/219.0f - 16.0f/219.0f;
    #endif
    #ifdef _rgb
    float2 coordIn = norm(coord, inImage);
    float3 rgb = read_imagef(inImage, sampler, coordIn).rgb;
    float  luma    = dot((float3)(0.2126f, 0.7152f, 0.0722f), rgb);
    #endif

    ushort   indx_x, indx_y, indx_z, indx_w;
    indx_x = convert_ushort_sat(rgb.x * 255.0f) + 127;
    indx_y = convert_ushort_sat(rgb.y * 255.0f) + 127;
    indx_z = convert_ushort_sat(rgb.z * 255.0f) + 127;
    indx_w = convert_ushort_sat(luma  * 255.0f) + 127;
    atom_inc(&tmp_histogram[indx_x]);
    atom_inc(&tmp_histogram[NUM_BINS+(uint)indx_y]);
    atom_inc(&tmp_histogram[2*NUM_BINS+(uint)indx_z]);
    atom_inc(&tmp_histogram[3*NUM_BINS+(uint)indx_w]);

    barrier(CLK_LOCAL_MEM_FENCE);

    // copy the partial histogram to appropriate location in histogram given by group_indx
    if (local_size >= (NUM_BINS * NUM_CHANNELS))
    {
        if (tid < (NUM_BINS * NUM_CHANNELS))
            histogram_buffer[group_indx + tid] = tmp_histogram[tid];
    }
    else
    {
        j = NUM_BINS * NUM_CHANNELS;
        indx = 0;
        do 
        {
            if (tid < j)
                histogram_buffer[group_indx + indx + tid] = tmp_histogram[indx + tid];
                
            j -= local_size;
            indx += local_size;
        } while (j > 0);
    }
}
