/**
 * Takes one or more input images/planes and outputs a single rgb texture after applying yuv -> rgb conversion
 * if necessary.
 */
in      vec2          texCoord0;
#ifdef FMT_NV12
in      vec2          texCoord1;
#endif

#ifdef FMT_YUV_PLANAR
in      vec2          texCoord1;
in      vec2          texCoord2;
#endif

out     vec4          outColor;

uniform mat3  coef;
uniform float lumaZero;
uniform float chromaZero;
uniform float blackLevel;

uniform vec2  offset;
uniform vec2  scale;

// Deinterlace
uniform int   deinterlaceMode;

// Zebra
uniform int   zebraEnabled;
const   float yUpper = 1.0 - (1.0 * 2.2 / 255.0);
const   float yLower = 0.0 + (1.0 * 2.2 / 255.0);
const   vec3  toY = vec3(0.2126, 0.7152, 0.0722);
const   vec3  lineColorOver = vec3(1.0, 0.0, 0.0);
const   vec3  lineColorUnder = vec3(0.0, 0.0, 1.0);
const   float lineSpacing = 8.0;
const   vec3  bgColorOver = vec3(0.5, 0.0, 0.0);
const   vec3  bgColorUnder = vec3(0.0, 0.0, 0.5);

uniform sampler2DRect inPlane0;
#ifdef FMT_NV12
uniform sampler2DRect inPlane1;
#endif
#ifdef FMT_YUV_PLANAR
uniform sampler2DRect inPlane1;
uniform sampler2DRect inPlane2;
#endif

vec4 textureDeinterlace(sampler2DRect image, vec2 pos) {
  int mod = int(mod(pos.y + deinterlaceMode, 2));
  return mix(texture(image, vec2(pos.x, floor(pos.y) - mod)), texture(image, vec2(pos.x, floor(pos.y) + mod)), 0.5);
}

vec3 zebra(vec3 rgb) {
  float y = dot(toY, rgb);

  if (y >= yUpper) {
    float modulo = mod(texCoord0.x + texCoord0.y, lineSpacing);
    float dist = abs(lineSpacing / 2.0 - modulo);
    float step = smoothstep(0.0, lineSpacing / 2.0 - 1.0, dist);
    return mix(lineColorOver, bgColorOver, step);

  } else if (y <= yLower) {
    float modulo = mod(texCoord0.x - texCoord0.y, lineSpacing);
    float dist = abs(lineSpacing / 2.0 - modulo);
    float step = smoothstep(0.0, lineSpacing / 2.0 - 1.0, dist);
    return mix(lineColorUnder, bgColorUnder, step);
  }

  return rgb;
}

#ifdef FMT_NV12
void main() {
  vec3 yuv;
  yuv.x  = deinterlaceMode == -1 ? texture(inPlane0,  texCoord0).r  : textureDeinterlace(inPlane0, texCoord0).r;
  yuv.yz = deinterlaceMode == -1 ? texture(inPlane1,  texCoord1).rg : textureDeinterlace(inPlane1, texCoord1).rg;
  vec3 rgb = coef * (yuv - vec3(lumaZero, chromaZero, chromaZero)) + vec3(blackLevel, blackLevel, blackLevel);
  rgb = zebraEnabled == 1 ? zebra(rgb) : rgb;
  outColor = vec4(rgb, 1.0);
}
#endif
#ifdef FMT_YUV_PLANAR
void main() {
  vec3 yuv;

  yuv.x = (deinterlaceMode == -1 ? texture(inPlane0, texCoord0).r : textureDeinterlace(inPlane0, texCoord0).r) * DEPTH_SCALE;
  yuv.y = (deinterlaceMode == -1 ? texture(inPlane1, texCoord1).r : textureDeinterlace(inPlane1, texCoord1).r) * DEPTH_SCALE;
  yuv.z = (deinterlaceMode == -1 ? texture(inPlane2, texCoord2).r : textureDeinterlace(inPlane2, texCoord2).r) * DEPTH_SCALE;

  vec3 rgb = coef * (yuv - vec3(lumaZero, chromaZero, chromaZero)) + vec3(blackLevel, blackLevel, blackLevel);
  rgb = zebraEnabled == 1 ? zebra(rgb) : rgb;
  outColor = vec4(rgb, 1.0);
}
#endif
#ifdef FMT_UYVY
void main() {
  // Note the order: gbr instead of rgb. See https://www.opengl.org/registry/specs/APPLE/rgb_422.txt
  vec3 yuv = deinterlaceMode == -1 ? texture(inPlane0, texCoord0).gbr : textureDeinterlace(inPlane0, texCoord0).gbr;
  vec3 rgb = coef * (yuv - vec3(lumaZero, chromaZero, chromaZero)) + vec3(blackLevel, blackLevel, blackLevel);
  rgb = zebraEnabled == 1 ? zebra(rgb) : rgb;
  outColor = vec4(rgb, 1.0);
}
#endif
#ifdef FMT_ARGB
void main() {
  vec4 rgba   = deinterlaceMode == -1 ? texture(inPlane0, texCoord0).yxwz : textureDeinterlace(inPlane0, texCoord0).yxwz;
  vec3 rgb    = zebraEnabled == 1 ? zebra(vec3(rgba.r, rgba.g, rgba.b)) : vec3(rgba.r, rgba.g, rgba.b);
  outColor    = vec4(rgb, rgba.a);
}
#endif
#ifdef FMT_RGBA
void main() {
  vec4 rgba   = deinterlaceMode == -1 ? texture(inPlane0, texCoord0).zyxw : textureDeinterlace(inPlane0, texCoord0).zyxw;
  vec3 rgb    = zebraEnabled == 1 ? zebra(vec3(rgba.r, rgba.g, rgba.b)) : vec3(rgba.r, rgba.g, rgba.b);
  outColor    = vec4(rgb, rgba.a);
}
#endif
#ifdef FMT_BGRA
void main() {
  vec4 rgba   = deinterlaceMode == -1 ? texture(inPlane0, texCoord0).xyzw : textureDeinterlace(inPlane0, texCoord0).xyzw;
  vec3 rgb    = zebraEnabled == 1 ? zebra(vec3(rgba.r, rgba.g, rgba.b)) : vec3(rgba.r, rgba.g, rgba.b);
  outColor    = vec4(rgb, rgba.a);
}
#endif
#ifdef FMT_RGB24
void main() {
  vec3 rgb    = deinterlaceMode == -1 ? texture(inPlane0, texCoord0).xyz : textureDeinterlace(inPlane0, texCoord0).xyz;
       rgb    = zebraEnabled == 1 ? zebra(rgb) : rgb;
  outColor    = vec4(rgb, 1.0);
}
#endif
